from typing import Literal, Tuple, Any, Dict, Optional

import hydra
import omegaconf
import rul_datasets
import pytorch_lightning as pl
from torch import nn

from rul_adapt.approach import ConsistencyApproach, SupervisedApproach


def get_consistency_dann(
    dataset: Literal["cmapss", "xjtu-sy"],
    source_fd: int,
    target_fd: int,
    pre_trainer_kwargs: Optional[Dict[str, Any]] = None,
    trainer_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[
    Tuple[rul_datasets.RulDataModule, SupervisedApproach, pl.Trainer],
    Tuple[
        rul_datasets.DomainAdaptionDataModule,
        nn.Module,
        ConsistencyApproach,
        pl.Trainer,
    ],
]:
    """
    Construct a Consistency DANN approach with the original hyperparameters.

    Examples:
        ```pycon
        >>> import rul_adapt
        >>> pre, main = rul_adapt.construct.get_consistency_dann("cmapss", 3, 1)
        >>> pre_dm, pre_approach, pre_trainer = pre
        >>> dm, approach, domain_disc, trainer = main
        ```

    Args:
        dataset: The name of the dataset, either `cmapss` or `xjtu-sy`.
        source_fd: The source FD of CMAPSS.
        target_fd: The target FD of CMAPSS.
        pre_trainer_kwargs: Overrides for the pre-training trainer class.
        trainer_kwargs: Overrides for the main trainer class.
    Returns:
        pre: The data module, approach and trainer for the pre-training stage
        main: The data module, approach, domain discriminator and trainer for
              the main stage
    """
    config = get_consistency_dann_config(dataset, source_fd, target_fd)
    consistency_dann = consistency_dann_from_config(
        config, pre_trainer_kwargs, trainer_kwargs
    )

    return consistency_dann


def get_consistency_dann_config(
    dataset: Literal["cmapss", "xjtu-sy"], source_fd: int, target_fd: int
) -> omegaconf.DictConfig:
    """
    Get a configuration for the Consistency DANN approach.

    The configuration can be modified and fed to [consistency_dann_from_config]
    [rul_adapt.construct.consistency.consistency_dann_from_config] to create the
    approach.

    Args:
        dataset: The name of the dataset, either `cmapss` or `xjtu-sy`.
        source_fd: The source FD of CMAPSS.
        target_fd: The target FD of CMAPSS.
    Returns:
        The Consistency DANN configuration.
    """
    _validate(dataset, source_fd, target_fd)
    with hydra.initialize("config", version_base="1.1"):
        config = hydra.compose(
            "base",
            overrides=[
                f"+dataset={dataset}",
                f"dm.source.fd={source_fd}",
                f"dm.target.fd={target_fd}",
            ],
        )

    return config


def consistency_dann_from_config(
    config: omegaconf.DictConfig,
    pre_trainer_kwargs: Optional[Dict[str, Any]] = None,
    trainer_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[
    Tuple[rul_datasets.RulDataModule, SupervisedApproach, pl.Trainer],
    Tuple[
        rul_datasets.DomainAdaptionDataModule,
        nn.Module,
        ConsistencyApproach,
        pl.Trainer,
    ],
]:
    """
    Construct a Consistency DANN approach from a configuration.
    The configuration can be created by calling [get_consistency_dann_config]
    [rul_adapt.construct.consistency.get_consistency_dann_config].

    Args:
        config: The Consistency DANN config.
        pre_trainer_kwargs: Overrides for the pre-training trainer class.
        trainer_kwargs: Overrides for the main trainer class.
    Returns:
        pre: The data module, approach and trainer for the pre-training stage
        main: The data module, approach, domain discriminator and trainer for
              the main stage
    """
    source = hydra.utils.instantiate(config.dm.source)
    target = source.get_compatible(**config.dm.target)
    kwargs = hydra.utils.instantiate(config.dm.kwargs)
    dm_pre = rul_datasets.RulDataModule(source, **kwargs)
    dm = rul_datasets.DomainAdaptionDataModule(
        dm_pre, rul_datasets.RulDataModule(target, **kwargs)
    )

    feature_extractor = hydra.utils.instantiate(config.feature_extractor)
    regressor = hydra.utils.instantiate(config.regressor)
    domain_disc = hydra.utils.instantiate(config.domain_disc)

    approach_pre = hydra.utils.instantiate(config.consistency_pre)
    approach_pre.set_model(feature_extractor, regressor)

    approach = hydra.utils.instantiate(config.consistency)

    pre_trainer_kwargs = pre_trainer_kwargs or {}
    trainer_kwargs = trainer_kwargs or {}
    trainer_pre = hydra.utils.instantiate(config.trainer_pre, **pre_trainer_kwargs)
    trainer = hydra.utils.instantiate(config.trainer, **trainer_kwargs)

    return (dm_pre, approach_pre, trainer_pre), (dm, approach, domain_disc, trainer)


def _validate(
    dataset: Literal["cmapss", "xjtu-sy"], source_fd: int, target_fd: int
) -> None:
    if dataset not in ["cmapss", "xjtu-sy"]:
        raise ValueError(f"No configuration for '{dataset}'.")
    elif source_fd == target_fd:
        raise ValueError(
            f"No configuration for adapting from FD{source_fd:03} to itself."
        )
    elif dataset == "cmapss":
        _validate_cmapss(source_fd, target_fd)
    elif dataset == "xjtu-sy":
        _validate_xjtu_sy(source_fd, target_fd)


def _validate_cmapss(source_fd: int, target_fd: int):
    if 1 > source_fd or source_fd > 4:
        raise ValueError(f"CMAPSS has only FD001 to FD004 but no FD{source_fd:03}")
    elif 1 > target_fd or target_fd > 4:
        raise ValueError(f"CMAPSS has only FD001 to FD004 but no FD{target_fd:03}")


def _validate_xjtu_sy(source_fd: int, target_fd: int):
    if 1 > source_fd or source_fd > 3:
        raise ValueError(f"XJTU-SY has only FD001 to FD003 but no FD{source_fd:03}")
    elif 1 > target_fd or target_fd > 3:
        raise ValueError(f"XJTU-SY has only FD001 to FD003 but no FD{target_fd:03}")



"""The Consistency DANN approach uses a consistency loss in tandem with a [DANN
loss][rul_adapt.approach.dann]. First, the network is pre-trained in a supervised
fashion on the source domain. The pre-trained weights are then used to initialize the
main training stage. The consistency loss penalizes the weights of the feature
extractor moving away from the pre-trained version. This way the feature extractor
weights stay close to the pre-trained weights.

```python
# pre-training stage

Source --> PreFeatEx --> Source Feats --> Regressor  --> RUL Prediction

# main training stage

   ------- PreTrainFeatEx --> PreTrain Source Feats --> Consistency Loss
   |
   |
Source --> FeatEx --> Source Feats -----------> Regressor  --> RUL Prediction
        ^         |                 |
        |         |                 v
Target --         --> Target Feats -->  GRL --> DomainDisc --> Domain Prediction
```

This version of DANN was introduced by
[Siahpour et al.](https://doi.org/10.1109/TIM.2022.3162283)."""

import copy
import math
from itertools import chain
from typing import Optional, Any, List, Tuple, Literal

import numpy as np
import torch
from torch import nn

import rul_adapt.loss
from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import AdaptionEvaluator
from rul_adapt.model import FullyConnectedHead


class ConsistencyApproach(AdaptionApproach):
    """The Consistency DANN approach introduces a consistency loss that keeps the
    weights of the feature extractor close to the ones of a pre-trained version. This
    approach should only be used with a pre-trained feature extractor. Otherwise,
    the consistency loss would serve no purpose.

    The regressor and domain discriminator need the same number of input units as the
    feature extractor has output units. The discriminator is not allowed to have an
    activation function on its last layer for it to work with the DANN loss.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> disc = model.FullyConnectedHead(16, [8, 1], act_func_on_last_layer=False)
        >>> pre = approach.SupervisedApproach("rmse")
        >>> pre.set_model(feat_ex, reg, disc)
        >>> main = approach.ConsistencyApproach(1.0, 100)
        >>> main.set_model(pre.feature_extractor, pre.regressor, disc)

    """

    CHECKPOINT_MODELS = ["dann_loss", "frozen_feature_extractor"]

    dann_loss: rul_adapt.loss.DomainAdversarialLoss
    frozen_feature_extractor: nn.Module

    def __init__(
        self,
        consistency_factor: float,
        max_epochs: int,
        loss_type: Literal["mse", "mae", "rmse"] = "rmse",
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a new consistency DANN approach.

        The consistency factor is the strength of the consistency loss' influence.
        The influence of the DANN loss is increased during the training process. It
        starts at zero and reaches one at `max_epochs`.

        The domain discriminator is set by the `set_model` function together with the
        feature extractor and regressor. For more information, see the [approach]
        [rul_adapt.approach] module page.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            consistency_factor: The strength of the consistency loss' influence.
            max_epochs: The number of epochs after which the DANN loss' influence is
                        maximal.
            loss_type: The type of regression loss, either 'mse', 'rmse' or 'mae'.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.consistency_factor = consistency_factor
        self.max_epochs = max_epochs
        self.loss_type = loss_type
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        self.train_source_loss = utils.get_loss(loss_type)
        self.consistency_loss = rul_adapt.loss.ConsistencyLoss()
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)
        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    def set_model(
        self,
        feature_extractor: nn.Module,
        regressor: nn.Module,
        domain_disc: Optional[nn.Module] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Set the feature extractor, regressor and domain discriminator for this approach.
        The discriminator is not allowed to have an activation function on its last
        layer and needs to use only a single output neuron. It is wrapped by a
        [DomainAdversarialLoss][rul_adapt.loss.DomainAdversarialLoss].

        A frozen copy of the feature extractor is produced to be used for the
        consistency loss. The feature extractor should, therefore, be pre-trained.

        Args:
            feature_extractor: The pre-trained feature extraction network.
            regressor: The optionally pre-trained RUL regression network.
            domain_disc: The domain discriminator network.
        """
        domain_disc = self._check_domain_disc(domain_disc)
        super().set_model(feature_extractor, regressor, *args, **kwargs)
        self.dann_loss = rul_adapt.loss.DomainAdversarialLoss(domain_disc)
        self.frozen_feature_extractor = copy.deepcopy(feature_extractor)
        self.frozen_feature_extractor.requires_grad_(False)  # freeze network
        self.log_model_hyperparameters("domain_disc")

    def _check_domain_disc(self, domain_disc: Optional[nn.Module]) -> nn.Module:
        if domain_disc is None:
            raise ValueError(
                "No domain discriminator was set. This approach is unlikely to work."
            )
        if (
            isinstance(domain_disc, FullyConnectedHead)
            and domain_disc.act_func_on_last_layer
        ):
            raise ValueError(
                "Domain discriminator has an activation function on its last layer. "
                "This is not allowed due to torch.nn.BCEWithLogitsLoss being used as "
                "its loss. Please set 'act_func_on_last_layer' to False."
            )

        return domain_disc

    @property
    def domain_disc(self):
        """The domain discriminator network."""
        if hasattr(self, "dann_loss"):
            return self.dann_loss.domain_disc
        else:
            raise RuntimeError("Domain disc used before 'set_model' was called.")

    @property
    def dann_factor(self):
        """
        Return the influency of the DANN loss based on the current epoch.

        It is calculated as: `2 / (1 + math.exp(-10 * current_epoch / max_epochs)) - 1`
        """
        return 2 / (1 + math.exp(-10 * self.current_epoch / self.max_epochs)) - 1

    def configure_optimizers(self) -> utils.OptimizerLRSchedulerConfig:
        """Configure an optimizer to train the feature extractor, regressor and
        domain discriminator."""
        parameters = chain(
            self.feature_extractor.parameters(),
            self.regressor.parameters(),
            self.dann_loss.parameters(),
        )  # excludes frozen_feature_extractor from optimization
        optim = self._get_optimizer(parameters)

        return optim

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        return self.regressor(self.feature_extractor(inputs))

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of three tensors representing the source
        features, source labels and target features. Both types of features are fed
        to the feature extractor. Then the regression loss for the source domain and
        the DANN loss between domains is computed. Afterwards the consistency loss is
        calculated. The regression, DANN, consistency and combined loss are logged.

        Args:
            batch: A list of a source feature, source label and target feature tensors.
            batch_idx: The index of the current batch.
        Returns:
            The combined loss.
        """
        source, source_labels, target = batch
        source_labels = source_labels[:, None]

        frozen_source = self.frozen_feature_extractor(source)
        source = self.feature_extractor(source)
        target = self.feature_extractor(target)

        rul_preds = self.regressor(source)
        rmse_loss = self.train_source_loss(rul_preds, source_labels)

        dann_loss = self.dann_loss(source, target)
        consistency_loss = self.consistency_loss(frozen_source, source)

        loss = (
            rmse_loss
            + self.dann_factor * dann_loss
            + self.consistency_factor * consistency_loss
        )

        self.log("train/loss", loss)
        self.log("train/source_loss", self.train_source_loss)
        self.log("train/dann", self.dann_loss)
        self.log("train/consistency", self.consistency_loss)

        return loss

    def on_train_epoch_start(self) -> None:
        self.log("train/dann_factor", self.dann_factor)

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.
        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.
        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.
        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.
        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)


class StdExtractor:
    """
    This extractor can be used to extract the per-feature standard deviation from
    windows of data. It can be used to pre-process datasets like FEMTO and XJTU-SY
    with the help of the [RulDataModule][rul_datasets.core.RulDataModule].

    Examples:
        Extract the std of the horizontal acceleration and produce windows of size 30.
        ```pycon
        >>> import rul_datasets
        >>> import rul_adapt
        >>> fd1 = rul_datasets.XjtuSyReader(fd=1)
        >>> extractor = rul_adapt.approach.consistency.StdExtractor([0])
        >>> dm = rul_datasets.RulDataModule(fd1, 32, extractor, window_size=30)
        ```
    """

    def __init__(self, channels: List[int]) -> None:
        """
        Create a new feature extractor for standard deviations.

        Args:
            channels: The list of channel indices to extract features from.
        """
        self.channels = channels

    def __call__(
        self, inputs: np.ndarray, targets: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Extract features from the input data.

        The input is expected to have a shape of `[num_windows, window_size,
        num_features]`. The output will have a shape of `[num_windows,
        len(self.channels)]`.

        Args:
            inputs: The input data.
        Returns:
            The features extracted from the input data.
        """
        return np.std(inputs[:, :, self.channels], axis=1), targets


class TumblingWindowExtractor:
    def __init__(self, window_size: int, channels: List[int]) -> None:
        self.window_size = window_size
        self.channels = channels

    def __call__(
        self, features: np.ndarray, targets: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        org_window_size = features.shape[1]
        window_multiplier = org_window_size // self.window_size
        crop_size = self.window_size * window_multiplier
        num_channels = len(self.channels)

        features = features[:, :, self.channels]
        features = features[:, :crop_size].reshape(-1, self.window_size, num_channels)
        targets = np.repeat(targets, window_multiplier)

        return features, targets



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rul_adapt',
 'rul_adapt.approach',
 'rul_adapt.construct',
 'rul_adapt.construct.adarul',
 'rul_adapt.construct.cnn_dann',
 'rul_adapt.construct.consistency',
 'rul_adapt.construct.latent_align',
 'rul_adapt.construct.lstm_dann',
 'rul_adapt.construct.tbigru',
 'rul_adapt.loss',
 'rul_adapt.model']

package_data = \
{'': ['*'],
 'rul_adapt.construct.adarul': ['config/*', 'config/task_source/*'],
 'rul_adapt.construct.cnn_dann': ['config/*'],
 'rul_adapt.construct.consistency': ['config/*', 'config/dataset/*'],
 'rul_adapt.construct.latent_align': ['config/*',
                                      'config/dataset/*',
                                      'config/split_steps/*',
                                      'config/subtask/*'],
 'rul_adapt.construct.lstm_dann': ['config/*', 'config/task/*'],
 'rul_adapt.construct.tbigru': ['config/*', 'config/task/*']}

install_requires = \
['dtaidistance>=2.3.10,<3.0.0',
 'hydra-core>=1.3.1,<2.0.0',
 'pytorch-lightning>1.8.0.post1',
 'pywavelets>=1.4.1,<2.0.0',
 'rul-datasets>=0.14.1',
 'tqdm>=4.62.2,<5.0.0']

setup_kwargs = {
    'name': 'rul-adapt',
    'version': '0.5.0',
    'description': 'A collection of unsupervised domain adaption approaches for RUL estimation.',
    'long_description': '# RUL Adapt\n\n[![Master](https://github.com/tilman151/rul-adapt/actions/workflows/on_push.yaml/badge.svg)](https://github.com/tilman151/rul-adapt/actions/workflows/on_push.yaml)\n[![Release](https://github.com/tilman151/rul-adapt/actions/workflows/on_release.yaml/badge.svg)](https://github.com/tilman151/rul-adapt/actions/workflows/on_release.yaml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nThis library contains a collection of unsupervised domain adaption algorithms for RUL estimation.\nThey are provided as [LightningModules](https://pytorch-lightning.readthedocs.io/en/stable/api/lightning.pytorch.core.LightningModule.html#lightning.pytorch.core.LightningModule) to be used in [PyTorch Lightning](https://pytorch-lightning.readthedocs.io/en/latest/).\n\nCurrently, five approaches are implemented, including their original hyperparameters:\n\n* **LSTM-DANN** by Da Costa et al. (2020)\n* **ADARUL** by Ragab et al. (2020)\n* **LatentAlign** by Zhang et al. (2021)\n* **TBiGRU** by Cao et al. (2021)\n* **Consistency-DANN** by Siahpour et al. (2022)\n\nThree approaches are implemented without their original hyperparameters:\n\n* **ConditionalDANN** by Cheng et al. (2021)\n* **ConditionalMMD** by Cheng et al. (2021)\n* **PseudoLabels** as used by Wang et al. (2022)\n\nThis includes the following general approaches adapted for RUL estimation:\n\n* **Domain Adaption Neural Networks (DANN)** by Ganin et al. (2016)\n* **Multi-Kernel Maximum Mean Discrepancy (MMD)** by Long et al. (2015)\n\nEach approach has an example notebook which can be found in the [examples](https://github.com/tilman151/rul-adapt/tree/master/examples) folder.\n\n## Installation\n\nThis library is pip-installable. Simply type:\n\n```bash\npip install rul-adapt\n```\n\n## Contribution\n\nContributions are always welcome. Whether you want to fix a bug, add a feature or a new approach, just open an issue and a PR.',
    'author': 'Krokotsch, Tilman',
    'author_email': 'tilman.krokotsch@tu-berlin.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://krokotsch.eu/rul-adapt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""The Conditional Adaption approaches are derived from the [MMD] [
rul_adapt.approach.mmd] and [DANN][rul_adapt.approach.dann] approaches. They apply
their respective adaption loss not only to the whole data but also separately to
subsets of the data with a [ConditionalAdaptionLoss]
[rul_adapt.loss.conditional.ConditionalAdaptionLoss]. Fuzzy sets with rectangular
membership functions define these subsets.

Both variants were introduced by
[Cheng et al.](https://doi.org/10.1007/s10845-021-01814-y) in 2021."""

from copy import deepcopy
from typing import List, Tuple, Literal, Optional, Any

import torch
from torch import nn

import rul_adapt
from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import AdaptionEvaluator
from rul_adapt.model import FullyConnectedHead
from rul_adapt.utils import OptimizerLRSchedulerConfig


class ConditionalMmdApproach(AdaptionApproach):
    """The conditional MMD uses a combination of a marginal and conditional MML loss
    to adapt a feature extractor to be used with the source regressor.

    The regressor needs the same number of input units as the feature extractor has
    output units.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> cond_mmd = approach.ConditionalMmdApproach(0.01, 5, 0.5, [(0, 1)])
        >>> cond_mmd.set_model(feat_ex, reg)
    """

    def __init__(
        self,
        mmd_factor: float,
        num_mmd_kernels: int,
        dynamic_adaptive_factor: float,
        fuzzy_sets: List[Tuple[float, float]],
        loss_type: Literal["mse", "rmse", "mae"] = "mae",
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a new conditional MMD approach.

        The strength of the influence of the MMD loss on the feature extractor is
        controlled by the `mmd_factor`. The higher it is, the stronger the influence.
        The dynamic adaptive factor controls the balance between the marginal MMD and
        conditional MMD losses.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            mmd_factor: The strength of the MMD loss' influence.
            num_mmd_kernels: The number of kernels for the MMD loss.
            dynamic_adaptive_factor: The balance between marginal and conditional MMD.
            fuzzy_sets: The fuzzy sets for the conditional MMD loss.
            loss_type: The type of regression loss, either 'mse', 'rmse' or 'mae'.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.mmd_factor = mmd_factor
        self.num_mmd_kernels = num_mmd_kernels
        self.dynamic_adaptive_factor = dynamic_adaptive_factor
        self.loss_type = loss_type
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)

        # training metrics
        self.train_source_loss = utils.get_loss(self.loss_type)
        self.mmd_loss = rul_adapt.loss.MaximumMeanDiscrepancyLoss(self.num_mmd_kernels)
        conditional_mmd_losses = [
            rul_adapt.loss.MaximumMeanDiscrepancyLoss(self.num_mmd_kernels)
            for _ in range(len(fuzzy_sets))
        ]
        self.conditional_mmd_loss = rul_adapt.loss.ConditionalAdaptionLoss(
            conditional_mmd_losses, fuzzy_sets, mean_over_sets=True
        )

        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    @property
    def fuzzy_sets(self) -> List[Tuple[float, float]]:
        return self.conditional_mmd_loss.fuzzy_sets

    def configure_optimizers(self) -> OptimizerLRSchedulerConfig:
        """Configure an Adam optimizer."""
        return self._get_optimizer(self.parameters())

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        return self.regressor(self.feature_extractor(inputs))

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of three tensors representing the source
        features, source labels and target features. Both types of features are fed
        to the feature extractor. Then the regression loss for the source domain,
        the MMD loss and the conditional MMD loss are computed. The
        regression, MMD, conditional MMD and combined loss are logged.

        Args:
            batch: A list of a source feature, source label and target feature tensors.
            batch_idx: The index of the current batch.
        Returns:
            The combined loss.
        """
        source, source_labels, target = batch
        source_labels = source_labels[:, None]
        daf = self.dynamic_adaptive_factor

        source = self.feature_extractor(source)
        target = self.feature_extractor(target)
        source_preds = self.regressor(source)
        target_preds = self.regressor(target)

        source_loss = self.train_source_loss(source_preds, source_labels)
        mmd_loss = self.mmd_loss(source, target)
        cond_mmd_loss = self.conditional_mmd_loss(
            source, source_preds, target, target_preds
        )
        combined_mmd_loss = (1 - daf) * mmd_loss + daf * cond_mmd_loss
        loss = source_loss + self.mmd_factor * combined_mmd_loss

        self.log("train/loss", loss)
        self.log("train/source_loss", self.train_source_loss)
        self.log("train/mmd", self.mmd_loss)
        self.log("train/conditional_mmd", self.conditional_mmd_loss)

        return loss

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)


class ConditionalDannApproach(AdaptionApproach):
    """The conditional DANN approach uses a marginal and several conditional domain
    discriminators. The features are produced by a shared feature extractor. The loss
    in the domain discriminators is binary cross-entropy.

    The regressor and domain discriminators need the same number of input units as the
    feature extractor has output units. The discriminators are not allowed to have an
    activation function on their last layer and need to use only a single output
    neuron because [BCEWithLogitsLoss][torch.nn.BCEWithLogitsLoss] is used.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> disc = model.FullyConnectedHead(16, [8, 1], act_func_on_last_layer=False)
        >>> cond_dann = approach.ConditionalDannApproach(1.0, 0.5, [(0, 1)])
        >>> cond_dann.set_model(feat_ex, reg, disc)
    """

    CHECKPOINT_MODELS = ["dann_loss", "conditional_dann_loss"]

    dann_loss: rul_adapt.loss.DomainAdversarialLoss
    conditional_dann_loss: rul_adapt.loss.ConditionalAdaptionLoss

    def __init__(
        self,
        dann_factor: float,
        dynamic_adaptive_factor: float,
        fuzzy_sets: List[Tuple[float, float]],
        loss_type: Literal["mse", "rmse", "mae"] = "mae",
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a new conditional DANN approach.

        The strength of the domain discriminator's influence on the feature extractor
        is controlled by the `dann_factor`. The higher it is, the stronger the
        influence. The `dynamic_adaptive_factor` controls the balance between the
        marginal and conditional DANN loss.

        The domain discriminator is set by the `set_model` function together with the
        feature extractor and regressor. For more information, see the [approach]
        [rul_adapt.approach] module page.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            dann_factor: Strength of the domain DANN loss.
            dynamic_adaptive_factor: Balance between the marginal and conditional DANN
                                     loss.
            fuzzy_sets: Fuzzy sets for the conditional DANN loss.
            loss_type: The type of regression loss, either 'mse', 'rmse' or 'mae'.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.dann_factor = dann_factor
        self.dynamic_adaptive_factor = dynamic_adaptive_factor
        self.loss_type = loss_type
        self._fuzzy_sets = fuzzy_sets
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        self.train_source_loss = utils.get_loss(self.loss_type)
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)
        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    @property
    def fuzzy_sets(self) -> List[Tuple[float, float]]:
        return self._fuzzy_sets

    def set_model(
        self,
        feature_extractor: nn.Module,
        regressor: nn.Module,
        domain_disc: Optional[nn.Module] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Set the feature extractor, regressor, and domain discriminator for this
        approach.

        The discriminator is not allowed to have an activation function on its last
        layer and needs to use only a single output neuron.
        It is wrapped by a
        [DomainAdversarialLoss][rul_adapt.loss.DomainAdversarialLoss].

        A copy of the discriminator is used for each conditional loss governing a
        fuzzy set.

        Args:
            feature_extractor: The feature extraction network.
            regressor: The RUL regression network.
            domain_disc: The domain discriminator network.
                         Copied for each fuzzy set.
        """
        domain_disc = self._check_domain_disc(domain_disc)
        super().set_model(feature_extractor, regressor, *args, **kwargs)
        self.dann_loss = rul_adapt.loss.DomainAdversarialLoss(domain_disc)
        cond_losses = [deepcopy(self.dann_loss) for _ in range(len(self.fuzzy_sets))]
        self.conditional_dann_loss = rul_adapt.loss.ConditionalAdaptionLoss(
            cond_losses, self.fuzzy_sets
        )
        self.log_model_hyperparameters("domain_disc")

    def _check_domain_disc(self, domain_disc: Optional[nn.Module]) -> nn.Module:
        if domain_disc is None:
            raise ValueError(
                "No domain discriminator was set. This approach is unlikely to work."
            )
        if (
            isinstance(domain_disc, FullyConnectedHead)
            and domain_disc.act_func_on_last_layer
        ):
            raise ValueError(
                "Domain discriminator has an activation function on its last layer. "
                "This is not allowed due to torch.nn.BCEWithLogitsLoss being used as "
                "its loss. Please set 'act_func_on_last_layer' to False."
            )

        return domain_disc

    @property
    def domain_disc(self) -> nn.Module:
        return self.dann_loss.domain_disc

    def configure_optimizers(self) -> utils.OptimizerLRSchedulerConfig:
        """Configure an Adam optimizer."""
        return self._get_optimizer(self.parameters())

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        return self.regressor(self.feature_extractor(inputs))

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of three tensors representing the source
        features, source labels and target features. Both types of features are fed
        to the feature extractor. Then the regression loss for the source domain,
        the MMD loss and the conditional MMD loss are computed. The
        regression, MMD, conditional MMD and combined loss are logged.

        Args:
            batch: A list of a source feature, source label and target feature tensors.
            batch_idx: The index of the current batch.
        Returns:
            The combined loss.
        """
        source, source_labels, target = batch
        source_labels = source_labels[:, None]
        daf = self.dynamic_adaptive_factor

        source = self.feature_extractor(source)
        target = self.feature_extractor(target)
        source_preds = self.regressor(source)
        target_preds = self.regressor(target)

        source_loss = self.train_source_loss(source_preds, source_labels)
        dann_loss = self.dann_loss(source, target)
        cond_dann_loss = self.conditional_dann_loss(
            source, source_preds, target, target_preds
        )
        combined_dann_loss = (1 - daf) * dann_loss + daf * cond_dann_loss
        loss = source_loss + self.dann_factor * combined_dann_loss

        self.log("train/loss", loss)
        self.log("train/source_loss", self.train_source_loss)
        self.log("train/dann", self.dann_loss)
        self.log("train/conditional_dann", self.conditional_dann_loss)

        return loss

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for dataloader idx zero
        are assumed to be from the source domain and for dataloader idx one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)



"""The Adversarial Domain Adaption for Remaining Useful Life (ADARUL) approach
pre-trains a feature extractor and regressor on the source domain in a supervised
fashion. Afterwards the feature extractor is adapted by feeding it the target
features and training it adversarial against a domain discriminator. The
discriminator is trained to distinguish the source features fed to a frozen version
of the pre-trained feature extractor and the target features fed to the adapted
feature extractor.

The approach was first introduced by [Ragab et al.](
https://doi.org/10.1109/ICPHM49022.2020.9187053) and evaluated on the CMAPSS dataset."""

import copy
from typing import Optional, Any, List, Literal

import torch
from torch import nn

from rul_adapt import utils
from rul_adapt.approach.abstract import AdaptionApproach
from rul_adapt.approach.evaluation import AdaptionEvaluator
from rul_adapt.model import FullyConnectedHead


class AdaRulApproach(AdaptionApproach):
    """The ADARUL approach uses a GAN setup to adapt a feature extractor. This
    approach should only be used with a pre-trained feature extractor.

    The regressor and domain discriminator need the same number of input units as the
    feature extractor has output units. The discriminator is not allowed to have an
    activation function on its last layer for it to work with its loss.

    Examples:
        >>> from rul_adapt import model
        >>> from rul_adapt import approach
        >>> feat_ex = model.CnnExtractor(1, [16, 16, 1], 10, fc_units=16)
        >>> reg = model.FullyConnectedHead(16, [1])
        >>> disc = model.FullyConnectedHead(16, [8, 1], act_func_on_last_layer=False)
        >>> pre = approach.SupervisedApproach("mse", 125, lr=0.001)
        >>> pre.set_model(feat_ex, reg)
        >>> main = approach.AdaRulApproach(5, 1, 125, lr=0.001)
        >>> main.set_model(pre.feature_extractor, pre.regressor, disc)
    """

    CHECKPOINT_MODELS = ["_domain_disc", "frozen_feature_extractor"]

    _domain_disc: nn.Module
    frozen_feature_extractor: nn.Module

    def __init__(
        self,
        num_disc_updates: int,
        num_gen_updates: int,
        max_rul: Optional[int] = None,
        rul_score_mode: Literal["phm08", "phm12"] = "phm08",
        evaluate_degraded_only: bool = False,
        **optim_kwargs: Any,
    ) -> None:
        """
        Create a new ADARUL approach.

        The discriminator is first trained for `num_disc_updates` batches.
        Afterward, the feature extractor (generator) is trained for
        `num_gen_updates`. This cycle repeats until the epoch ends.

        The regressor is supposed to output a value between [0, 1] which is then
        scaled by `max_rul`.

        For more information about the possible optimizer keyword arguments,
        see [here][rul_adapt.utils.OptimizerFactory].

        Args:
            max_rul: Maximum RUL value of the training data.
            num_disc_updates: Number of batches to update discriminator with.
            num_gen_updates: Number of batches to update generator with.
            rul_score_mode: The mode for the val and test RUL score, either 'phm08'
                            or 'phm12'.
            evaluate_degraded_only: Whether to only evaluate the RUL score on degraded
                                    samples.
            **optim_kwargs: Keyword arguments for the optimizer, e.g. learning rate.
        """
        super().__init__()

        self.automatic_optimization = False  # use manual optimization loop

        self.num_disc_updates = num_disc_updates
        self.num_gen_updates = num_gen_updates
        self.max_rul = max_rul
        self.rul_score_mode = rul_score_mode
        self.evaluate_degraded_only = evaluate_degraded_only
        self.optim_kwargs = optim_kwargs

        self._disc_counter, self._gen_counter = 0, 0
        self._get_optimizer = utils.OptimizerFactory(**self.optim_kwargs)

        self.gan_loss = nn.BCEWithLogitsLoss()

        self.evaluator = AdaptionEvaluator(
            self.forward, self.log, self.rul_score_mode, self.evaluate_degraded_only
        )

        self.save_hyperparameters()

    def set_model(
        self,
        feature_extractor: nn.Module,
        regressor: nn.Module,
        domain_disc: Optional[nn.Module] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Set the feature extractor, regressor and domain discriminator for this approach.
        The discriminator is not allowed to have an activation function on its last
        layer and needs to use only a single output neuron.

        A frozen copy of the feature extractor is produced to be used for the *real*
        samples fed to the discriminator. The feature extractor should, therefore,
        be pre-trained.

        Args:
            feature_extractor: The feature extraction network.
            regressor: The RUL regression network.
            domain_disc: The domain discriminator network.
        """
        domain_disc = self._check_domain_disc(domain_disc)
        super().set_model(feature_extractor, regressor, *args, **kwargs)
        self._domain_disc = domain_disc
        self.frozen_feature_extractor = copy.deepcopy(feature_extractor)
        self.frozen_feature_extractor.requires_grad_(False)  # freeze network
        self.log_model_hyperparameters("_domain_disc")

    def _check_domain_disc(self, domain_disc: Optional[nn.Module]) -> nn.Module:
        if domain_disc is None:
            raise ValueError(
                "No domain discriminator was set. This approach is unlikely to work."
            )
        if (
            isinstance(domain_disc, FullyConnectedHead)
            and domain_disc.act_func_on_last_layer
        ):
            raise ValueError(
                "Domain discriminator has an activation function on its last layer. "
                "This is not allowed due to torch.nn.BCEWithLogitsLoss being used as "
                "its loss. Please set 'act_func_on_last_layer' to False."
            )

        return domain_disc

    @property
    def domain_disc(self):
        """The domain discriminator network."""
        if hasattr(self, "_domain_disc"):
            return self._domain_disc
        else:
            raise RuntimeError("Domain disc used before 'set_model' was called.")

    def configure_optimizers(self) -> List[utils.OptimizerLRSchedulerConfig]:  # type: ignore[override]
        """Configure an optimizer for the generator and discriminator respectively."""
        return [
            self._get_optimizer(self.domain_disc.parameters()),
            self._get_optimizer(self.feature_extractor.parameters()),
        ]

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Predict the RUL values for a batch of input features."""
        max_rul = self.max_rul or 1
        pred = self.regressor(self.feature_extractor(inputs)) * max_rul

        return pred

    def on_train_epoch_start(self) -> None:
        self._reset_update_counters()

    def training_step(self, batch: List[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Execute one training step.

        The `batch` argument is a list of three tensors representing the source
        features, source labels and target features. Each iteration either only the
        discriminator or only the generator is trained. The respective loss is logged.

        The *real* samples are source features passed though the frozen version of
        the feature extractor. The *fake* samples are the target features passed
        through the adapted feature extractor. The discriminator predicts if a sample
        came from the source or target domain.

        Args:
            batch: A list of a source feature, source label and target feature tensors.
            batch_idx: The index of the current batch.
        Returns:
            Either the discriminator or generator loss.
        """
        source, _, target = batch

        if self._updates_done():
            self._reset_update_counters()

        if self._should_update_disc():
            optim, _ = self.optimizers()  # type: ignore[attr-defined, misc]
            loss = self._get_disc_loss(source, target)
            self.log("train/disc_loss", loss)
            self._disc_counter += 1
        elif self._should_update_gen():
            _, optim = self.optimizers()  # type: ignore[attr-defined, misc]
            loss = self._get_gen_loss(target)
            self.log("train/gen_loss", loss)
            self._gen_counter += 1
        else:
            raise RuntimeError("Configuration error. Did update neither disc nor gen.")

        optim.zero_grad()  # type: ignore[union-attr]
        self.manual_backward(loss)
        optim.step()

        return loss

    def _should_update_disc(self):
        return self._disc_counter < self.num_disc_updates and self._gen_counter == 0

    def _should_update_gen(self):
        return (
            self._disc_counter == self.num_disc_updates
            and self._gen_counter < self.num_gen_updates
        )

    def _reset_update_counters(self):
        self._disc_counter, self._gen_counter = 0, 0

    def _updates_done(self) -> bool:
        return (
            self._disc_counter == self.num_disc_updates
            and self._gen_counter == self.num_gen_updates
        )

    def _get_disc_loss(self, source, target):
        batch_size = source.shape[0]
        source = self.frozen_feature_extractor(source).detach()
        target = self.feature_extractor(target).detach()
        domain_pred = self.domain_disc(torch.cat([source, target]))
        domain_labels = torch.cat(
            [
                torch.ones(batch_size, 1, device=self.device),  # real labels
                torch.zeros(batch_size, 1, device=self.device),  # fake labels
            ]
        )
        loss = self.gan_loss(domain_pred, domain_labels)

        return loss

    def _get_gen_loss(self, target):
        batch_size = target.shape[0]
        target = self.feature_extractor(target)
        domain_pred = self.domain_disc(target)
        domain_labels = torch.ones(batch_size, 1, device=self.device)  # flipped labels
        loss = self.gan_loss(domain_pred, domain_labels)

        return loss

    def validation_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one validation step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for `dataloader_idx` zero
        are assumed to be from the source domain and for `dataloader_idx` one from the
        target domain. The metrics are written to the configured logger under the
        prefix `val`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.validation(batch, domain)

    def test_step(
        self, batch: List[torch.Tensor], batch_idx: int, dataloader_idx: int
    ) -> None:
        """
        Execute one test step.

        The `batch` argument is a list of two tensors representing features and
        labels. A RUL prediction is made from the features and the validation RMSE
        and RUL score are calculated. The metrics recorded for `dataloader_idx` zero
        are assumed to be from the source domain and for `dataloader_idx` one from the
        target domain. The metrics are written to the configured logger under the
        prefix `test`.

        Args:
            batch: A list containing a feature and a label tensor.
            batch_idx: The index of the current batch.
            dataloader_idx: The index of the current dataloader (0: source, 1: target).
        """
        domain = utils.dataloader2domain(dataloader_idx)
        self.evaluator.test(batch, domain)



import torch
from torch import nn


class TwoStageExtractor(nn.Module):
    """This module combines two feature extractors into a single network.

    The input data is expected to be of shape `[batch_size, upper_seq_len,
    input_channels, lower_seq_len]`. An example would be vibration data recorded in
    spaced intervals, where lower_seq_len is the length of an interval and
    upper_seq_len is the window size of a sliding window over the intervals.

    The lower_stage is applied to each interval individually to extract features.
    The upper_stage is then applied to the extracted features of the window.
    The resulting feature vector should represent the window without the need to
    manually extract features from the raw data of the intervals.
    """

    def __init__(
        self,
        lower_stage: nn.Module,
        upper_stage: nn.Module,
    ):
        """
        Create a new two-stage extractor.

        The lower stage needs to take a tensor of shape `[batch_size, input_channels,
        seq_len]` and return a tensor of shape `[batch_size, lower_output_units]`. The
        upper stage needs to take a tensor of shape `[batch_size, upper_seq_len,
        lower_output_units]` and return a tensor of shape `[batch_size,
        upper_output_units]`. Args: lower_stage: upper_stage:
        """
        super().__init__()

        self.lower_stage = lower_stage
        self.upper_stage = upper_stage

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """
        Apply the two-stage extractor to the input tensor.

        The input tensor is expected to be of shape `[batch_size, upper_seq_len,
        input_channels, lower_seq_len]`. The output tensor will be of shape
        `[batch_size, upper_output_units]`.

        Args:
            inputs: the input tensor

        Returns:
            an output tensor of shape `[batch_size, upper_output_units]`
        """
        batch_size, upper_seq_len, input_channels, lower_seq_len = inputs.shape
        inputs = inputs.reshape(-1, input_channels, lower_seq_len)
        inputs = self.lower_stage(inputs)
        inputs = inputs.reshape(batch_size, upper_seq_len, -1)
        inputs = torch.transpose(inputs, 1, 2)
        inputs = self.upper_stage(inputs)

        return inputs



from functools import partial
from typing import cast

from torch import nn

from rul_adapt.model import CnnExtractor
from rul_adapt.model.wrapper import DropoutPrefix


def init_weights(feature_extractor: CnnExtractor, regressor: DropoutPrefix) -> None:
    """
    Initialize the weights of the feature extractor and regressor in-place.

    For the weight matrices the Xavier uniform initialization is used. The biases are
    initialized to zero. This function works only for the networks returned by a call to
    [rul_adapt.construct.get_cnn_dann][].

    Args:
        feature_extractor: The feature extractor network to be initialized.
        regressor: The regressor network to be initialized.
    """
    feature_extractor.apply(_weight_init_fe)
    layer_count = 0
    for m in regressor.modules():
        if layer_count < len(cast(list, regressor.wrapped.units)) - 1:
            m.apply(_weight_init_fc_tanh)
            layer_count += 1
        else:
            m.apply(_weight_init_fc_linear)


def _weight_init_fe(m):
    if isinstance(m, nn.Conv1d) or isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight, gain=nn.init.calculate_gain("tanh"))
        nn.init.zeros_(m.bias)


def _weight_init_fc(m, gain):
    if isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight, gain=gain)
        nn.init.zeros_(m.bias)


_weight_init_fc_tanh = partial(_weight_init_fc, gain=nn.init.calculate_gain("tanh"))
_weight_init_fc_linear = partial(_weight_init_fc, gain=nn.init.calculate_gain("linear"))
